#include <asm/io.h>

#define	CLIO_BASE	0x03B0
#define	CLIO_REGION	0x03E0-CLIO_BASE

/*\
|*| Write Value to MISC Register & set Color or Mono
\*/
static inline void
outbMISC ( unsigned char value )
{
	outb ( value | MISC_COLOR_IO, MISCw ) ;	/* Write to misc Register */
}

/*\
|*| Read Byte from MISC Register
\*/
static inline unsigned char
inbMISC ( void )
{
	return inb ( MISCr ) ;
}

/*\
|*| Output Byte to 'reg' Register of CRT
\*/
static inline void
outbCRT ( unsigned char value, unsigned char reg )
{
	outw ( ( value << 8 ) | reg, CRTIc ) ;
}

/*\
|*| Input Byte from 'reg' Register of CRT
\*/
static inline unsigned char
inbCRT ( unsigned char reg )
{
	outb ( reg, CRTIc ) ;		/* program register to get */
	return inb ( CRTDc ) ;		/* get the value */
}

/*\
|*| Output Byte to 'reg' Register of SEQ
\*/
static inline void
outbSEQ ( unsigned char value, unsigned char reg )
{
	outw ( ( value << 8 ) | reg, SEQI ) ;
}

/*\
|*| Input Byte from 'reg' Register of SEQ
\*/
static inline unsigned char
inbSEQ ( unsigned char reg )
{
	outb ( reg, SEQI ) ;		/* program the register to get */
	return inb ( SEQD ) ;		/* get the value */
}

/*\
|*| Output Byte to 'reg' Register of GRC
\*/
static inline void
outbGRC ( unsigned char value, unsigned char reg )
{
	outw ( ( value << 8 ) | reg, GRCI ) ;
}

static inline void
outwGRC ( unsigned short value, unsigned char reg )
{
        outw ( ( value << 8 )     | (reg+0), GRCI ) ;
	outw ( ( value & 0xff00 ) | (reg+1), GRCI ) ;
}

static inline void
out3GRC ( unsigned int value, unsigned char reg )
{
        outw ( (value << 8 )        | (reg+0), GRCI ) ;
        outw ( (value & 0xff00)     | (reg+1), GRCI ) ; 
        outw ( ((value>>8)& 0xff00) | (reg+2), GRCI ) ;
}



/*\
|*| Input Byte from 'reg' Register of GRC
\*/
static inline unsigned char
inbGRC ( unsigned char reg )
{
	outb ( reg, GRCI ) ;		/* program the register to get */
	return inb ( GRCD ) ;		/* get the value */
}

/*\
|*| Output Byte to 'reg' Register of ATC
\*/
static inline void
outbATC ( unsigned char value, unsigned char reg )
{
	outw ( ( value << 8 ) | reg, ATCI ) ;
}

/*\
|*| Input Byte from 'reg' Register of ATC
\*/
static inline unsigned char
inbATC ( unsigned char reg )
{
	inb ( INS1rc ) ;
	outb ( reg, ATCI ) ;		/* program the register to get */
	return inb ( ATCDr ) ;		/* get the value */
}

/* routines for Hidden DAC register */

static inline void
outbHDR(unsigned char value)
{
	inb(0x03C6);inb(0x03C6); /* read 4 times */
	inb(0x03C6);inb(0x03C6);
	outb(value,0x03C6); /* Write value */
}

static inline unsigned char
inbHDR(void)
{
	unsigned char hdr;

	inb(0x03C6);inb(0x03C6); /* read 4 times */
	inb(0x03C6);inb(0x03C6);
	hdr = inb(0x03C6); /* read value */
	inb(0x03C7); /* reset counter */
	return hdr;
}

void
cl_lock ( void )
{
	outbSEQ ( 0x0F, 0x06 ) ;
}

void
cl_unlock ( void )
{
int	temp ;

	outbSEQ ( 0x12, 0x06 ) ;

	temp = inbCRT ( CRT_VSYNCEND ) ;
	outbCRT ( CRT_VSYNCEND, temp & ( CR11_VSYNCEND_MASK || CR11_CLEAR_VSYNC_IRQ || CR11_DISABLE_VSYNC_IRQ || CR11_5_DRAMCYCLES ) ) ;
}


#define NoSEQRegs	0x20
#define	NoCRTRegs	0x1C
#define NoGRCRegs	0x0F
#define NoATCRegs	0x15

struct statetable
{
   unsigned char MISC;
   unsigned char FCTRL;
   unsigned char HDR;
   unsigned char SEQ[NoSEQRegs];
   unsigned char CRT[NoCRTRegs];
   unsigned char GRC[NoGRCRegs];
   unsigned char ATC[NoATCRegs];
};

#ifdef NEED_SAVE_STATE
static void
save_SEQ(unsigned char *buf)
{
  int i;

  for(i = 0; i < NoSEQRegs; i++)
   {
     outb(i, SEQI); *(buf++)=inb(SEQD);
     #ifdef DEBUG_IO
       if (!(i & 0xF))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  #ifdef DEBUG_IO
    printk("\n");
  #endif
}
#endif

static void
restore_SEQ(unsigned char *buf)
{
  register unsigned char i;

  outbSEQ(SR00_SYNCRESET,0);
  for(i = 1; i < 0x05; i++)
      outbSEQ(buf[i],i);
  outbSEQ((buf[0x07] & ~0x0F) | 0x01,0x07);
  outbSEQ(buf[0x0E],0x0E);
  outbSEQ(buf[0x1E],0x1E);
  outbSEQ(buf[0x07],0x07);
  outbSEQ(buf[0x0F],0x0F);
  outbSEQ(buf[0x16],0x16);
  outbSEQ(buf[0x17],0x17);
  outbSEQ(buf[0x1F],0x1F);
  outbSEQ(buf[0],0);
}

#ifdef NEED_SAVE_STATE
static void
save_CRT(unsigned char *buf)
{
  int i;

  for(i = 0; i < NoCRTRegs; i++)
   { 
     outb(i, CRTIc); *(buf++)=inb(CRTDc);
     #ifdef DEBUG_IO
       if (!(i & 0xF))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  #ifdef DEBUG_IO
    printk("\n");
  #endif
}
#endif

static void
restore_CRT(unsigned char buf[])
{
  register unsigned char i;

  outbCRT(inbCRT(0x11) & ~CR11_LOCKTIMING, 0x11);
  for(i = 0; i < NoCRTRegs; i++)
    outbCRT(buf[i], i); 
}

#ifdef NEED_SAVE_STATE
static void
save_GRC(unsigned char *buf)
{
  int i;

  for(i = 0; i < NoGRCRegs; i++) 
   {
     outb(i, GRCI); *(buf++)=inb(GRCD);
     #ifdef DEBUG_IO
       if (!(i & 0xF))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  #ifdef DEBUG_IO
    printk("\n");
  #endif
}
#endif

static void
restore_GRC(unsigned char *buf)
{
  register unsigned char i;

  for(i = 0; i < NoGRCRegs; i++)
    {
      outb(i, GRCI); outb(*(buf++), GRCD);
      #ifdef DEBUG_IO
        if(!(i & 0xF))
          printk("\n%.2x: ",i);
        printk("%.2X, ",*(buf-1));
      #endif
    }
  #ifdef DEBUG_IO
    printk("\n");
  #endif
}

#ifdef NEED_SAVE_STATE
static void
save_ATC(unsigned char *buf)
{
  int i;

  for(i = 0; i < NoATCRegs; i++)
   { 
     inb(INS1rc); outb(i, ATCI); /* reset AFF and setup index */
     *(buf++) = inb(ATCDr);
     #ifdef DEBUG_IO
       if (!(i & 0xF))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  inb(INS1rc); outb(ATCI_ENABLEDISPLAY, ATCI);
}
#endif

static void
restore_ATC(unsigned char *buf)
{
  register unsigned char i;

  for(i = 0; i < NoATCRegs; i++)
   { 
     inb(INS1rc); outb(i,ATCI); outb(*(buf++),ATCDw);
     #ifdef DEBUG_IO
       if(!(i & 0x0F))
         printk("\n%.2x: ",i);
       printk("%.2X, ",*(buf-1));
     #endif
   }
  inb(INS1rc); outb(ATCI_ENABLEDISPLAY, ATCI);
}

#ifdef NEED_SAVE_STATE
static void
save_state(struct statetable *state)
{
  state->MISC  = inbMISC();	/* save miscellaneous output reg */
  #ifdef DEBUG_IO
    printk("MISC: %.2X\n",state->MISC);
  #endif
  state->HDR = inbHDR();      /* save Hidden DAC */
  save_SEQ(state->SEQ);		/* sequencer regs */
  save_CRT(state->CRT);		/* cathode ray tube contr. regs */
  save_GRC(state->GRC);		/* graphics controller regs */
  save_ATC(state->ATC);		/* attribute controller regs */
}
#endif

static void
restore_state(struct statetable *state)
{
   outbMISC((inb(MISCr) & MISC_CLOCK_MASK) | state->MISC);
   outb(state->FCTRL, FCTRLwc);
   outbHDR(state->HDR);
   restore_SEQ(state->SEQ);
   restore_CRT(state->CRT);
   restore_GRC(state->GRC);
   restore_ATC(state->ATC);
}

#ifdef DEBUG_IO
static void
dump_state(struct statetable *state)
{ int i;
  printk("%.2x\n",state->MISC);
  for(i = 0; i < NoSEQRegs; i++)
   { if(!(i & 0xF)) printk("\n%.2x: ",i); printk("%.2x, ",state->SEQ[i]); }
  for(i = 0; i < NoCRTRegs; i++)
   { if(!(i & 0xf)) printk("\n%.2x: ",i); printk("%.2x, ",state->CRT[i]); }
  for(i = 0; i < NoGRCRegs; i++)
   { if(!(i & 0xF)) printk("\n%.2x: ",i); printk("%.2x, ",state->GRC[i]); }
  for(i = 0; i < NoATCRegs; i++)
   { if(!(i & 0xF)) printk("\n%.2x: ",i); printk("%.2x, ",state->ATC[i]); }
}
#endif
