/* 
   S3 generic Acceleration Driver - 29 Jul 1996

   Copyright (C) 1995 Andreas Beck - becka@sunserver1.rz.uni-duesseldorf.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "S3/accel.h"

/********************
 * Helper Functions *
 ********************/

int s3_wait_fifo(int spc_need)
{ int x=1000000000;
  while((inb(ECR_GP_STATUS)&(0x100>>spc_need))&&--x) schedule();
  return(!x);
}

static int s3_wait_idle(void)
{ int x=1000000000;
  while((inw(ECR_GP_STATUS) & EC06_GE_BUSY)&&--x) schedule();
  return(!x);
}

/*********************************************
**** Accelerated drawing functions follow ****
*********************************************/

static int s3_drawbox(struct kgi_state *gr)
{ struct box *boxp=(struct box *)_io_buf;

  if (   gr->mode.request.graphtype==GT_4BIT
      || gr->mode.request.graphtype==GT_TEXT )
    return(-ENODRVSUP_MODE_MMAP);

  if (boxp->w&&boxp->h)				/* 0 width is not OK ! */
  {
	if (s3_wait_fifo(3)) return(-ENODRVSUP_NOW_MMAP);
	outw(0x27,ECR_FOREGR_MIX);
	outl(get_fs_long(&gr->drawinfo->fg_color),ECR_FOREGROUND);
	outw(MF_PIX_CONTROL,		ECR_MULTIFUNC);

	if (s3_wait_fifo(5)) return(-ENODRVSUP_NOW_MMAP);
	outw(boxp->x,			ECR_CUR_X);
	outw(boxp->y,			ECR_CUR_Y);
	outw(boxp->w-1,			ECR_MAJAXIS_CNT);
	outw((boxp->h-1) | MF_MINAXIS_CNT,ECR_MULTIFUNC);
	outw(EC07_FILLRECT | 
	     EC07_LITTLE_ENDIAN |
	     EC07_32BIT_IO |
	     EC07_DIR(5) |
	     EC07_DRAWPIXEL |
	     EC07_ACROSS_PLANE |
             0x0001 ,ECR_COMMAND);
	s3_wait_idle();
  }
  return(0);
}

static int s3_copybox(struct kgi_state *gr)
{ struct cbox *boxp=(struct box *)_io_buf;
  int dir=0;

  if (   gr->mode.request.graphtype==GT_4BIT
      || gr->mode.request.graphtype==GT_TEXT )
    return(-ENODRVSUP_MODE_MMAP);

  if (boxp->w&&boxp->h)				/* 0 width is not OK ! */
  {
	if (s3_wait_fifo(2)) return(-ENODRVSUP_NOW_MMAP);
	outw(0x67,ECR_FOREGR_MIX);
	outw(MF_PIX_CONTROL,		ECR_MULTIFUNC);

	if (boxp->x>boxp->x2) {dir|=EC07_DIR(1);}
	else { boxp->x+=boxp->w-1;boxp->x2+=boxp->w-1;}
	if (boxp->y>boxp->y2) {dir|=EC07_DIR(4);}
	else { boxp->y+=boxp->h-1;boxp->y2+=boxp->h-1;}

	if (s3_wait_fifo(7)) return(-ENODRVSUP_NOW_MMAP);
	outw(boxp->x,			ECR_CUR_X);
	outw(boxp->y,			ECR_CUR_Y);
	outw(boxp->x2,			ECR_DIAG_STEP);
	outw(boxp->y2,			ECR_AXIAL_STEP);
	outw(boxp->w-1,			ECR_MAJAXIS_CNT);
	outw((boxp->h-1) | MF_MINAXIS_CNT,ECR_MULTIFUNC);
	outw(EC07_BITBLT | 
	     EC07_LITTLE_ENDIAN |
	     EC07_32BIT_IO |
	     dir |
	     EC07_DRAWPIXEL |
	     EC07_ACROSS_PLANE |
             0x0001 ,ECR_COMMAND);
	s3_wait_idle();
  }
  return(0);
}

static int s3_drawline(struct kgi_state *gr)
{ int dx,dy,min,max;
  struct line *linp=(struct line *)_io_buf;

  if (   gr->mode.request.graphtype==GT_4BIT
      || gr->mode.request.graphtype==GT_TEXT )
    return(-ENODRVSUP_MODE_MMAP);

  dx=linp->x2-linp->x;
  dy=linp->y2-linp->y;
  
  if (abs(dx)>abs(dy))	{max=abs(dx);min=abs(dy);}
  else 			{max=abs(dy);min=abs(dx);}

  if (max==0) return(0);	/* FIXME ! */

  if (s3_wait_fifo(3)) return(-ENODRVSUP_NOW_MMAP);

  outw(0x27,ECR_FOREGR_MIX);
  outl(get_fs_long(&gr->drawinfo->fg_color),ECR_FOREGROUND);
  outw(MF_PIX_CONTROL,		ECR_MULTIFUNC);

  if (s3_wait_fifo(7)) return(-ENODRVSUP_NOW_MMAP);
  outw(linp->x,			ECR_CUR_X);
  outw(linp->y,			ECR_CUR_Y);

  outw((max-1),			ECR_MAJAXIS_CNT);
  outw((min-max)<<1,		ECR_DIAG_STEP);
  outw((min<<1),		ECR_AXIAL_STEP);

  outw((min+min-max-(dx<0)),	ECR_ERROR_TERM);

  outw(EC07_LINE
        | ((dy>0)            ? EC07_DIR(4) : 0)
	| ((dx>0)            ? EC07_DIR(1) : 0)
	| ((abs(dy)>abs(dx)) ? EC07_DIR(2) : 0)
	| EC07_DRAWPIXEL 
	| EC07_ACROSS_PLANE
	| 0x0001	,ECR_COMMAND);

  s3_wait_idle();
  return(0);
}

static int sel_page=0;

int
kgi_graphics_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		   unsigned long arg)
{ struct kgi_state *gr = file->private_data;

  if (gr->graph_num != fg_console)	/* Don't accel inactive screen ! */
    return(-ENODRVSUP_MODE_MMAP);

  switch(cmd)
  { case GRAPHICS_DRAWPIXEL:
#if 0	/* 1=SLOW */
	{  struct ggi_coord *h = (void*) _io_buf;
	   unsigned offset = h->x + h->y*current_mode.virtual.x;
	   int col=0x12345678;

	    switch(current_mode.graphtype)
	     {
	       case GT_8BIT:	((char*) fb_graphics)[offset] = (char) col;
				return EOK;
			
	       case GT_15BIT:
	       case GT_16BIT:	((short *) fb_graphics)[offset] = (short) col;
				return EOK;

	       case GT_32BIT:	((int *) fb_graphics)[offset] = (int) col;
				return EOK;
	       default:		printk("DUNNO MYSELF ... ???");
	     }
	  } break;
#else
	     return -ENODRVSUP_ALWAYS_MMAP;
#endif
    case GRAPHICS_SET16COL:
	outbGRC((*(int *)_io_buf)& 3,4);       /* Read Map select */
	outbGRC((*(int *)_io_buf)&15,0);       /* Color select */
	return EOK;
    case GRAPHICS_SETPAGE: 
    	sel_page=*((int *)_io_buf);
	kgi_unmap_all();
	return EOK;
    case GRAPHICS_SETDRAWINFO:
	gr->drawinfo=*(ggi_drawinfo **)_io_buf;
	return EOK;

    /* Really accelerated functions. */
    case GRAPHICS_DRAWBOX  :
	return(s3_drawbox(gr));
    case GRAPHICS_COPYBOX  :
	return(s3_copybox(gr));
    case GRAPHICS_DRAWLINE :
	return(s3_drawline(gr));

    /* Functions that will probably benefit from lower-lever accel */
    case GRAPHICS_FILLSCREEN:
	return -ENODRVSUP_ALWAYS_LOWER;

    default: 
	TRACE(printk("Unknown graphics command %x.\n",cmd););
	return -ENODRVSUP_ALWAYS_MMAP;
    /* for functions that will probably fallback to something accelerated,
       you should return ALWAYS_LOWER instead */
  }
  return -ENODRVSUP_ALWAYS_CANT;	/* This driver can't do this. */
}
