/*\
|*| --------------------------------------------------------------------------
|*|     KGI kernel driver
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by 
|*|
|*|	  Andreas Beck,		[becka@sunserver1.rz.uni-duesseldorf.de]
|*|	  Steffen Seeger	[seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#define	MAINTAINER	"Andreas Beck, [becka@sunserver1.rz.uni-duesseldorf.de]"

/* This is the main part - thus no __NO_VERSION__ */
#include <linux/module.h>
#include <linux/version.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/types.h>
#include <linux/event.h>
#include <linux/sched.h>
#include <asm/pgtable.h>
#include <linux/console.h>

#include <ggi/mapping.h>

#ifdef DEBUG
#define	NEED_GRAPHTYPE_SYMS
#endif

/* FIXME ! Something in here breaks the struct event declaration ... ??? */
#include <ggi/graphdev.h>
#include <linux/tty.h>

#if STEFFEN
#define	VC_VFB_SIZE	(3*4096-sizeof(struct kgi_vc))	/* size of virtual fb */
#else
#define	VC_VFB_SIZE	(3*4096)	/* size of virtual fb */
#endif

void 	*fb_graphics;	/* phys. address of the graphics framebuffer  */
size_t	fb_graph_len;	/* size of the graphics framebuffer           */
void	*fb_text;		/* physical address of the text framebuffer   */
size_t	fb_text_len;	/* size of the text framebuffer		      */

static struct kgi_state *graph[MAX_NR_CONSOLES];

ggi_mode current_mode;


/*\
|*| Event handling and ramdaemon interface.
\*/

struct kernel_event *pending_ev=NULL;
struct kernel_event *old_ev=NULL;
int ramdaemon=-1;



#ifdef DEBUG

static void
printk_mode (struct kgi_mode *mode, char *msg)
{
  #define M(X) mode->X
  #define R(X) mode->request.X
  #define T(X) M(X.width), M(X.blankstart), M(X.syncstart), M(X.syncend),\
		M(X.blankend), M(X.total), M(X.polarity)
  printk(msg);
  printk("mode = {\n  {{ %i, %i},  /* visible */\n",
			R(visible.x),R(visible.y));
  printk("   { %i, %i},   /* virtual */\n",R(virtual.x),R(virtual.y));
  printk("   { %i, %i},   /* size [mm] */\n",R(size.x),R(size.y));
  printk("   %s,\n", GRAPHMASK(R(graphtype)));
  printk("   { %i, %i}},   /* textgrid */\n", R(textgrid.x),R(textgrid.y));
  printk("   { %i, %i, %i, %i, %i, %i, %i}, /* x timing */\n", T(x));
  printk("   { %i, %i, %i, %i, %i, %i, %i}, /* y timing */\n", T(y));
  printk("   { %i, %i},   /* magnify */\n", M(magnify.x), M(magnify.y));
  printk("   %x,          /* flags   */\n", M(flags));
  printk("   %i,          /* dclk    */\n", M(dclk));
  printk("   { %i, %i},   /* clk     */\n", M(clk.mul), M(clk.div));
  printk("   { %i, %i}    /* lclk    */\n}\n", M(lclk.mul), M(lclk.div));

  #undef M
  #undef R
  #undef T
}
#endif /* #ifdef DEBUG */

#define	MAX_EVENTS	16	/* *must* be a power of 2 !!	*/

enum kgi_strategy
{
   STR_DISCARD,	/* discard framebuffer, signal redraw request	*/
   STR_LAST	/* last defined					*/
};

struct kgi_mode defaultmode
		    = { { {640,400},
			  {640,400},
			  {0,0},
			  GT_TEXT,
			  {8,16} 
			},
		      }	;


static void
kgi_free_state(struct kgi_state *state)
{
  if (state)
    {
       if (state->font0.free_data)
         vfree(state->font0.data);

       if (state->font1.free_data)
         vfree(state->font1.data);

       vfree(state);
    }
}

static int
kgi_check_mode(struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int cnt = 10;

  if (cmd == CMD_PROPOSE)
    {
      memset(&(mode->x), 0, sizeof(struct kgi_mode) - sizeof(ggi_mode));
      TRACE(printk_mode(mode, "mode ok:\n"));
      switch(mode->request.graphtype)
       {
	 case GT_TEXT:			/* propose a standard text mode */
		mode->palette = &default_text_palette;
		mode->font0   = &default_font;
		mode->font1   = &default_font;

		if (   ((mode->request.virtual.x/mode->request.textgrid.x) 
		      *(mode->request.virtual.y/mode->request.textgrid.y)
		      *sizeof(u_short))
		    > VC_VFB_SIZE)
		  return -ENOMEM;
		break;

         case GT_1BIT: 
	 case GT_4BIT:			/* indexed colors, but no text font */
	 case GT_4BIT_PLANES:
	 case GT_8BIT:
		mode->palette  = &default_graphics_palette;
		mode->font0    = mode->font1 = NULL;
		break;

	 case GT_15BIT:			/* only gamma correction makes sense */
	 case GT_16BIT:			
	 case GT_24BIT:
	 case GT_32BIT:
		mode->palette  = &default_gamma_map;
		mode->font0    = mode->font1 = NULL;
	        break;

	 default:
		INTERNALERROR;
       }
    }

  while (cnt-- && !(cmd == CMD_READY))
   {
     register int err;
     TRACE(printk("check_mode: cmd = %i; cnt = %i\n", cmd, cnt));
     if ((err = kgi_clock_check_mode  (mode,  cmd)))  return err;
     TRACE(printk("check_mode: clock ok.\n"));	 
     if ((err = kgi_ramdac_check_mode (mode,  cmd)))  return err;
     TRACE(printk("check_mode: ramdac ok.\n"));
     if ((err = kgi_chipset_check_mode(mode,  cmd)))  return err;
     TRACE(printk("check_mode: chipset ok.\n"));
     if ((err = kgi_monitor_check_mode(mode, &cmd)))  return err;
     TRACE(printk("check_mode: monitor ok."));

     TRACE(printk_mode(mode, "mode:\n"));
   }

  TRACE(printk_mode(mode, "mode ok:\n"));
 
  return (++cnt) ? EOK : -EDRIVER;
}

static void
kgi_set_mode(struct kgi_mode *mode)
{
  if (memcmp(&current_mode, &(mode->request), sizeof(current_mode)))
    {					/* avoid some stress on hardware by  */
      kgi_chipset_set_mode (mode);	/* unneccessary mode initialisations */
      current_mode = mode->request;
    }
}

void *
kgi_get_physical_mem(unsigned long *fb, unsigned long start, unsigned long end,
                     unsigned long align, size_t size, char *device)
{
   /*\
   |*|	This needs a housekeeping of the physical RAM allocation not to
   |*|	conflict with other cards. Anyhow, for now let's Plug&Pray.
   \*/
   #define END_MEM 0x10000000		/* end of RAM (!!! FIXME!) */

   void *p = NULL;

   if ((0x000A0000 <= start) && (size < 0x60000))
     {
	*fb = start;
        return (void *) start;
     }

   start = (start < END_MEM) ? END_MEM : start;	/* don't touch RAM    */
   *fb = end & ~(align - 1);			/* align fb poperly   */

   while ((start < *fb) && !p)		/* try to 'allocate' physical */
    {					/* memory. We start from the  */
      p = vremap(*fb, size);		/* of the allowed range.      */
      *fb -= align; 
    }
   *fb += align;

   return p;
}

/*\
|*|	GGI driver interface
\*/

static int
kgi_driver_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		 unsigned long arg)
{
  struct kgi_state *gr = file->private_data;
  int err;
  int console = fg_console;

  TRACE(printk("IOCTL-driver started %x\n",cmd));
  switch(cmd)
   {
     case DRIVER_TESTDRIVER:	/* check if driver is present	*/
	return DEVICEFILE_MAGIC;

     case DRIVER_TESTMODE:		/* check if mode is possible	*/
	_out_buf = _io_buf;		/* _io_cnt is sizeof(ggi_mode)	*/
	err = kgi_check_mode((struct kgi_mode *) _io_buf, CMD_PROPOSE);
	return err;

     case DRIVER_SETGRAPHMODE:
	_out_buf = _io_buf;		/* _io_cnt is sizeof(ggi_mode)  */

	if ((err = kgi_check_mode((struct kgi_mode *) _io_buf, CMD_PROPOSE)))
	  return err;

	gr->mode = (*(struct kgi_mode*) _io_buf);
	gr->palette=*(gr->mode.palette);
	gr->mode.palette=&gr->palette;
	memcpy(gr->color,gr->palette.data,
	       sizeof(gr->color[0])*gr->palette.cnt);
	gr->palette.data=gr->color;
	
	if ((gr->graph_num == console))
	  kgi_set_mode(&(gr->mode));

	/* !!! buffer reallocation!  */
	return EOK;

     case DRIVER_GETGRAPHMODE:	/* return graphics mode		*/
	_out_buf = &(gr->mode.request);
	return EOK;			/* _io_cnt is sizeof(ggi_mode)	*/

     case DRIVER_RAMDAEMON_INSTALL:
	if (!suser()) return -EPERM;
	ramdaemon=current->pid;
	TRACE(printk("RAMDAEMON attached.\n"));
	return 0;
        	
     case DRIVER_RAMDAEMON_ASK:
	{ ramdaemon_shm *rsh=(ramdaemon_shm *)_io_buf;
	  _out_buf=_io_buf;
	  if (pending_ev)
	  { rsh->self=*((int *)pending_ev->data)+1;
	    TRACE(printk("RAMDAEMON asked for %d.\n",rsh->self));
	    rsh->len=2048*1024;		/* 2 MB */
	  }
	  else return -EINVAL;
	} 
	return 0;

     case DRIVER_RAMDAEMON_CPY:
	{ ramdaemon_shm *rsh=(ramdaemon_shm *)_io_buf;
	  TRACE(printk("RAMDAEMON copy %p\n",rsh->shmat));
	  if (pending_ev)
	  { switch(pending_ev->event)
	    { case E_CONS_VTREL:
	        /*CurrState=&StateNow[*((int *)pending_ev->data)+1];*/
		if (rsh->shmat)		/* There might have been problems ... */
	    		kgi_savescreen(rsh->shmat);	/* Do the copy !*/
	    	kgi_set_mode(&defaultmode);		/* Default textmode. */
		/* Fixme ! Locking ? */
            	break;
	            	
	      case E_CONS_VTACQ:
		/* Locking ? */
		kgi_set_mode(&(graph[*((int *)pending_ev->data)]->mode));
		if (rsh->shmat)		/* There might have been problems ... */
		    	kgi_loadscreen(rsh->shmat);	/* Do the copy !*/
            	break;

              default:break;
	    }
	    if (--pending_ev->lock_cnt==0 && pending_ev->callback)
	    { old_ev=pending_ev;pending_ev=NULL;old_ev->callback(old_ev); }
	  }
	}
	return 0;
     default:
	TRACE(printk("Unknown command: %x\n",cmd));
	return -EDRIVER;
   }
}

static void
kgi_done(void)
{
  kgi_clock_done();			/* deinitialisation, the calling     */
  kgi_ramdac_done();			/* sequence is reverse to initiali-  */
  kgi_monitor_done();			/* sation.			     */
  kgi_chipset_done();
}

static int
kgi_init(void)
{
  memset(&current_mode, 0, sizeof(current_mode));

  if (kgi_chipset_init())		/* chipset provides IO functions for */
    return -EIO;			/* all other parts, thus init first. */

  if (kgi_monitor_init())		/* chipset ok. Now init the monitor  */
    {					/* driver. If this fails (e.g. not   */
      kgi_chipset_done();		/* detected) deinitialize chipset.   */
      return -EIO;
    }
  if (kgi_ramdac_init())		/* Now go for the ramdac. 	     */
    {					
      kgi_monitor_done();		/* failed, thus deinitialize...	     */
      kgi_chipset_done();
      return -EIO;
    }
  if (kgi_clock_init())			/* last is the clock driver. 	     */
    {
      kgi_ramdac_done();		/* failed :-(			     */
      kgi_monitor_done();
      kgi_chipset_done();
      return -EIO;
    }

  return 0;
}

/*\
|*|	/dev/graph file operations
\*/


typedef int (*ioctl_fn)(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg); 

static ioctl_fn ioctl[8] =
{
  kgi_driver_ioctl,	/* NOTE: this depends on the IOCTL definitions ! */
  kgi_monitor_ioctl,
  kgi_ramdac_ioctl,
  kgi_clock_ioctl,
  kgi_chipset_ioctl,
  kgi_graphics_ioctl,
  NULL,
  NULL
};

unsigned int	_io_cnt;
unsigned char   _io_buf[_IO_BUF_SIZE];
         void	*_out_buf;

static int
graph_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
	  unsigned long arg)
{
  static int lock = 0;
  register int foo;
  int result = 0;

  if (lock) return -EBUSY;	/* ??? could this happen (new ioctl while  */ 
  lock = 1;			/* ??? the old one is not finished ?)	   */

  _out_buf = NULL;
  _io_cnt = _IOC_SIZE(cmd);

  TRACE(printk("IOCTL started %x\n",cmd));
  if (cmd & IOC_IN)		/* do we have input ?	*/
    {
      if ((foo = verify_area (VERIFY_READ, (void *) arg, _io_cnt)))
        {
          result = foo;
          goto unlock;
        }
      memcpy_fromfs (_io_buf, (void *) arg, _io_cnt);
    }

  foo = _IOC_TYPE(cmd);
  if ((foo < 8) && ioctl[foo])
    result = ioctl[foo] (inode, file, cmd, arg);
  else
  { if (foo==0xff) /* MULTI request */
    { for(foo=0;foo<8;foo++)
      { if (ioctl[foo] &&
	    (result = ioctl[foo] (inode, file, cmd, arg))==0)
	      goto multi_ok;
      }
      result = -EINVAL;  goto unlock;
    }
    result = -EINVAL;  goto unlock; 
  }
  multi_ok:
  TRACE(printk("IOCTL completed %x result=%x\n",cmd,result));

  if (cmd & IOC_OUT)
    {
      if ((foo = verify_area (VERIFY_WRITE, (void *) arg, _io_cnt)))
        {
           result = foo;
           goto unlock;
        }
      if (_out_buf && (0 < _io_cnt) && (_io_cnt < (IOCSIZE_MASK+1)))
      {  memcpy_tofs ((void *) arg, _out_buf, _io_cnt);
      }
       else
        result = -EDRIVER;
    }

  unlock:
    lock = 0;
  TRACE(printk("IOCTL returns %x result=%x\n",cmd,result));
    return result;
}

/*\
|*| Mmap stuff (c) 1996 Andreas Beck
\*/

struct vm_operations_struct Linear_vmops = {
	NULL,			/* open */
	kgi_linearmap_close,	/* close */
	NULL,			/* unmap */
	NULL,			/* protect */
	NULL,			/* sync */
	NULL,			/* advise */
	kgi_linearmap_nopage,	/* nopage */
	NULL,			/* wppage */
	NULL,			/* swapout */
	NULL,			/* swapin */
	};

struct vm_operations_struct Paged_vmops = {
	NULL,			/* open */
	kgi_pagedmap_close,		/* close */
	NULL,			/* unmap */
	NULL,			/* protect */
	NULL,			/* sync */
	NULL,			/* advise */
	kgi_pagedmap_nopage,	/* nopage */
	NULL,			/* wppage */
	NULL,			/* swapout */
	NULL,			/* swapin */
	};

int graph_mmap(struct inode * inode, struct file * file, struct vm_area_struct * vma)
{
	switch(vma->vm_offset)
	{ default:return -ENXIO;	/* We don't support queer mappings right now ... */
	  case MMAP_FRAME :vma->vm_end=vma->vm_start+0x10000;
	  case MMAP_LINEAR:
	  	  vma->vm_inode = inode;
	  	  vma->vm_flags |= VM_FASTHANDLER;
		  inode->i_count++;
		  /* Make sure paging tables are there ... */
	 	  zeromap_page_range(vma->vm_start,vma->vm_end - vma->vm_start,vma->vm_page_prot);
		  fast_unmap_page_range(vma->vm_mm,vma->vm_start,vma->vm_end - vma->vm_start);
		  break;
	}
	switch(vma->vm_offset)
	{ default:return -ENXIO;	/* We don't support queer mappings right now ... */
	  case MMAP_LINEAR:
	 	  vma->vm_ops   = &Linear_vmops;
	 	  break;
	  case MMAP_FRAME:
	 	  vma->vm_ops   = &Paged_vmops;
	 	  break;
	}
	vma->vm_offset= (unsigned long) file->private_data;
	return 0;
}

static int 
graph_select(struct inode *inode, struct file *file, int type, select_table *wait)
{
#if STEFFEN
  struct kgi_state *gr = file->private_data;

  switch (type)
   {

     case SEL_IN:	/* read access possible?	*/

	if (can_get_event(&gr->ev_buffer))
          return 1;
        select_wait(&gr->ev_buffer.wait, wait);
	return 0;

     case SEL_OUT:	/* write access possible?	*/
     default:
	return 0;
   }
#endif
   return 0;
}

static int 
graph_read(struct inode *inode, struct file *file, char *buf, int count)
{
#if STEFFEN
  struct kgi_state *gr = file->private_data;
  int write = gr->ev_buffer.write;
  int read  = gr->ev_buffer.read;
  int cnt = write - read;
  int err;

  while (! cnt)			/* if there is no input availiable now, */
    {				/* block the process until we have sth. */
      if (current->signal & ~current->blocked)
        return 0;

      current->state = TASK_INTERRUPTIBLE;
      current->timeout = jiffies + 10; /* ms */
      schedule();

      write = gr->ev_buffer.write;
      cnt = write - read;
    }

  count /= sizeof(ggi_event);
  if (count < cnt)		/* more availiable than requested ?	*/
    {
      cnt = count;
      write = read + cnt; 
    }
  if (!cnt)
    return 0;

  if ((err = verify_area (VERIFY_WRITE, buf, cnt*sizeof(ggi_event))))
    return err;    /* ??? is this check neccessary or does VFS already ? */

  if (write < gr->ev_buffer.size)	/* can we read all at once ?	*/
    {
      int written = cnt*sizeof(ggi_event);

      memcpy_tofs(buf, &(gr->event[read]), written);	/* read out events */
      gr->ev_buffer.read = write;

      inode->i_atime = CURRENT_TIME;

      return written;
    }
   else
    {
      int first  = sizeof(ggi_event)*(gr->ev_buffer.size - read);
      int second = sizeof(ggi_event)*(write - gr->ev_buffer.size);

      memcpy_tofs(buf, &(gr->event[read]), first);		/* read out events */
      memcpy_tofs(buf+first, &(gr->event[0]), second);
      gr->ev_buffer.write &= gr->ev_buffer.size - 1;
      gr->ev_buffer.read  = write & (gr->ev_buffer.size - 1);

      inode->i_atime = CURRENT_TIME;

      return first + second;
    }
#endif
  return 0;
}

static int
graph_open(struct inode *inode, struct file *file)
{
  struct tty_struct *tty = current->tty;
  int minor = MINOR(inode->i_rdev) ? MINOR(inode->i_rdev) : MINOR(tty->device);

  if (   ((tty->driver.type != TTY_DRIVER_TYPE_CONSOLE)
          && (!suser() || MINOR(inode->i_rdev) == 0))
      ||  MINOR(tty->device) != minor) /* !!! change_console depends on this */
    return -EPERM;	/* no way, the controlling tty must be local!	*/

  minor--;

  if (graph[minor])	/* for now we allow only one process per VG	*/
    return -EBUSY;

  if (minor < MAX_NR_CONSOLES)
    {
      struct kgi_state *gr = vmalloc(sizeof(struct kgi_state));

      if (! gr)
        return -ENOMEM;

      file->private_data = graph[minor] = gr;

      memset(gr, 0, sizeof(*gr));

      gr->graph_num = minor;
      gr->mode      = defaultmode;		/* default to textmode	*/

      MOD_INC_USE_COUNT;
      return EOK;
    }
  return -ENODEV;
}

static void
graph_close(struct inode *inode, struct file *file)
{
  struct kgi_state *gr = file->private_data;
  int fg = fg_console;

  MOD_DEC_USE_COUNT;

  kgi_unmap_all(); 	/* Is this necessary ??? */
  if (fg==gr->graph_num) kgi_set_mode(&defaultmode);
  graph[gr->graph_num] = NULL;
  kgi_free_state(gr);
}

static struct file_operations 
kgi_fops =
{ NULL,		/* lseek	*/
  graph_read,	/* read		*/
  NULL,         /* write	*/
  NULL,         /* readdir	*/
  graph_select,	/* select	*/
  graph_ioctl,	/* ioctl	*/ 
  graph_mmap,	/* mmap		*/
  graph_open,	/* open		*/
  graph_close,	/* release	*/
  NULL		/* fsync	*/
};

/*\
|*| Event system
\*/

static int kgi_event(struct kernel_event *k_event)
{ int conso; 

  switch(k_event->event)
  { case E_CONS_SAK : /* We restore std text mode here ... */
#ifdef DEBUG
	printk("SAK called.\n");
#endif
	kgi_set_mode(&defaultmode);
  	break;

    case E_CONS_VTREL:		/* Console is released */
	if ((conso=*((int *)k_event->data)+1)<MAX_NR_CONSOLES && conso>0
	    && graph[conso-1]!=NULL) /* Is it under GGI control ? */
	 {	/* Yes. Schedule Event for later and tell RAMDAEMON */
		pending_ev=k_event;
		kgi_unmap_all();	/* Unmap all mapped pages. */
		/* Check for running ramdaemon while notifying. */
		if (kill_proc(ramdaemon,SIGUSR1,1)==0)
			return 1;
		else
		{ /* We have no ramdaemon. Ignore silently and set default text mode. */
		  #ifdef DEBUG
		    printk("KGI: ramdaemon not running - graphics not saved !\n");
		  #endif
		  kgi_set_mode(&defaultmode);
		  return 0; }
	}
    	break;

    case E_CONS_VTACQ:		/* Console is acquired */
    	if ((conso=*((int *)k_event->data)+1)<MAX_NR_CONSOLES && conso>0
    	    &&graph[conso-1]!=NULL)  /* Is it under GGI control ? */
	{	/* Yes. Schedule Event for later, set mode and tell RAMDAEMON */
		struct kgi_state *gr = graph[conso-1];
		
		pending_ev=k_event;
		kgi_unmap_all();	/* FIXME probably unneeded. */
		kgi_set_mode(&gr->mode);
		if (kill_proc(ramdaemon,SIGUSR1,1)==0)
			return 1;
		else
		{ /* We have no ramdaemon. So we can't restore contents. */
		  #ifdef DEBUG
		    printk("KGI: ramdaemon not running - graphics not restored !\n");
		  #endif
		  return 0; }
	}
    	break;
    case E_CONS_VTRESIZE:
	{ struct kgi_mode chkmode = defaultmode;
	  chkmode.request.visible.x=*(((int *)k_event->data)+0)* 8;
	  chkmode.request.visible.y=*(((int *)k_event->data)+1)*16;
	  chkmode.request.virtual.x=*(((int *)k_event->data)+0)* 8;
	  chkmode.request.virtual.y=*(((int *)k_event->data)+1)*16;
	  chkmode.request.textgrid.x= 8;
	  chkmode.request.textgrid.y=16;
	  if (kgi_check_mode(&chkmode , CMD_PROPOSE))
	    return 0;	/* We should take care for this ... */
	  defaultmode=chkmode;
	  kgi_set_mode(&defaultmode);
	  return 0;
	}
       break;
    default: break;
  }

  return 0;
}

static struct kernel_event_operations graphdev_evops =
{ NULL,
  kgi_event,
  NULL		/* We cannot cancel it ! */
};

/*\
|*|	module loading
\*/


int
init_module(void)
{
   if (kgi_init())
     return -EIO;

   if (register_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME, &kgi_fops))
     {
       printk("%s: can't register device file %s (major = %i).\n",\
	      __FILE__, DEVICEFILE_NAME, DEVICEFILE_MAJOR);
       return -EIO;
     }

   register_evhand(&graphdev_evops);	/* Activate the event handling */

   if (kgi_check_mode(&defaultmode, CMD_PROPOSE)) 
     printk("Check default mode failed ...\n");
   printk("kgi module loaded.\n");
   return 0;
}

void
cleanup_module (void)
{
  if (MOD_IN_USE)
    printk("kgi: kgi module is still in use. (not removed)\n");

  if (unregister_evhand(&graphdev_evops))
    printk("graphdev.c: unregister_evhand() failed\n");
  if (unregister_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME))
    printk("kgi: %s:%i: cleanup_module() failed.\n", __FILE__, __LINE__);
   else
    {
      kgi_done();
      printk("kgi module removed.\n");
    }

  /* !!! we have to restore console state... (difficult) ! */
}

