/*\
|*| --------------------------------------------------------------------------
|*|	SC 15025 / SC15026 ramdac driver
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1996, Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|	This code was derived from
|*|	[1]	Sierra Semiconductor Standard Products Databook, Sept 1992
\*/

#define	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
#define	DAC_VERSION	"0.01"

#define	SC1502x_125	0		/* 125 MHz version	*/
#define	SC1502x_110	1		/* 110 MHz version	*/
#define	SC1502x_80	2		/*  80 MHz version	*/
#define	SC1502x_66	3		/*  66 MHz verison	*/

#ifndef SC_VERISON
 #define	SC_VERSION	SC1502x_110	/* default version to assume  */
#endif

/* ---------------- no user serviceable code below this line ---------------- */

#ifdef SC15026
  #define	SC1502x	"SC 15026"
#else
  #define	SC1502x	"SC 15025"
#endif

#ifdef DEBUG
#define	NEED_GRAPHTYPE_SYMS
#endif

#include <ggi/graphdev.h>

#define	DAC_PRECISION	8	/* we can do 8 bit per RGB value	*/

#include "SC/1502x.h"
#include "IBM/clut.inc"

#if !((MONITORVERSION == 01) && (MODEVERSION == 03) && (RAMDACVERSION == 01))
  #warning kgi datastructures have changed, check this code !!
#endif

#define	SUPPORTED (GM_TEXT | GM_4BIT | GM_8BIT | GM_15BIT | GM_16BIT | \
                   GM_24BIT | GM_32BIT)

static struct kgi_ramdac 
dacinfo[4] =
{
	{
		"Sierra Semiconductor",		/* manufacturer		*/
	 	SC1502x " - 125",		/* model		*/
 		SUPPORTED,			/* supported modes	*/
	 	FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
		FL_DAC_2XCLOCK,
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_8BIT_PB,  			/* device type		*/
		  {       0, 125000000 },	/* normal mode		*/
		  {       0,  80000000 },	/* 2xclk mode		*/

		 {{	  0, 125000000 },	/* 	GT_TEXT		*/
		  {       0, 125000000 },	/* 	GT_1BIT		*/
		  {	  0, 125000000 },	/* 	GT_4BIT		*/
		  {       0, 125000000 },	/* 	GT_8BIT		*/
		  {	  0,  80000000 },	/* 	GT_15BIT	*/
		  { 	  0,  80000000 },	/* 	GT_16BIT	*/
		  {       0,  41666666 },	/* 	GT_24BIT	*/
		  {	  0,  40000000 },	/* 	GT_32BIT	*/
		  {	  0, 125000000 }}	/* 	GT_4BIT_PLANES	*/
        },
	{
		"Sierra Semiconductor",		/* manufacturer		*/
	 	SC1502x " - 110",		/* model		*/
 		SUPPORTED,			/* supported modes	*/
	 	FL_DAC_8BIT | FL_DAC_GAMMA |  	/* flags		*/
		FL_DAC_2XCLOCK,
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_8BIT_PB,  			/* device type		*/
		  {       0, 110000000 },	/* normal mode		*/
		  {       0,  65000000 },	/* 2xclk mode		*/

		 {{	  0, 110000000 },	/* 	GT_TEXT		*/
		  {       0, 110000000 },	/* 	GT_1BIT		*/
		  {	  0, 110000000 },	/* 	GT_4BIT		*/
		  {       0, 110000000 },	/* 	GT_8BIT		*/
		  {	  0,  65000000 },	/* 	GT_15BIT	*/
		  { 	  0,  65000000 },	/* 	GT_16BIT	*/
		  {       0,  36666666 },	/* 	GT_24BIT	*/
		  {	  0,  32500000 }	/* 	GT_32BIT	*/
		  {	  0, 110000000 }}	/* 	GT_4BIT_PLANES	*/
        },
	{
		"Sierra Semiconductor",		/* manufacturer		*/
	 	SC1502x " - 80",		/* model		*/
 		SUPPORTED,			/* supported modes	*/
	 	FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
		FL_DAC_2XCLOCK,
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_8BIT_PB,  			/* device type		*/
		  {       0,  80000000 },	/* normal mode		*/
		  {       0,  50000000 },	/* 2xclk mode		*/

		 {{	  0,  80000000 },	/* 	GT_TEXT		*/
		  {       0,  80000000 },	/* 	GT_1BIT		*/
		  {	  0,  80000000 },	/* 	GT_4BIT		*/
		  {       0,  80000000 },	/* 	GT_8BIT		*/
		  {	  0,  50000000 },	/* 	GT_15BIT	*/
		  { 	  0,  50000000 },	/* 	GT_16BIT	*/
		  {       0,  26666666 },	/* 	GT_24BIT	*/
		  {	  0,  25500000 },	/* 	GT_32BIT	*/
		  {	  0,  80000000 }}	/* 	GT_4BIT_PLANES	*/
        },
	{
		"Sierra Semiconductor",		/* manufacturer		*/
	 	SC1502x " - 66",		/* model		*/
 		SUPPORTED,			/* supported modes	*/
	 	FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
		FL_DAC_2XCLOCK,
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_8BIT_PB,  			/* device type		*/
		  {       0,  66000000 },	/* normal mode		*/
		  {       0,  66000000 },	/* 2xclk mode		*/

		 {{	  0,  66000000 },	/* 	GT_TEXT		*/
		  {       0,  66000000 },	/* 	GT_1BIT		*/
		  {	  0,  66000000 },	/* 	GT_4BIT		*/
		  {       0,  66000000 },	/* 	GT_8BIT		*/
		  {	  0,  50000000 },	/* 	GT_15BIT	*/
		  { 	  0,  50000000 },	/* 	GT_16BIT	*/
		  {       0,  22000000 },	/* 	GT_24BIT	*/
		  {	  0,  25000000 },	/* 	GT_32BIT	*/
		  {	  0,  66000000 }}	/* 	GT_4BIT_PLANES	*/
         }
};
static struct kgi_ramdac *ramdac = dacinfo + SC_VERSION;

/*\
|*|	I/O routines to access indexed extended registers.
|*|	NOTE: ERPF must be set to one before calling inbEXT, outbEXT !
\*/

static inline uchar
inbEXT(uchar index)
{
  outbDAC(index, DAC_E_INDEX_R);
  return inbDAC(DAC_E_DATA);
}

static void
outbEXT(uchar value, uchar index)
{
  outbDAC(index, DAC_E_INDEX_W);
  outbDAC(value, DAC_E_DATA); 
}

static void
outbDAC_CMD(uchar value)
{
  int i;

  inbDAC(DAC_PW_INDEX);	/* reset the IPF		*/

  for (i = 0; i < 4; i++)      /* set the IPF                  */
    inbDAC(DAC_PIXEL_MASK);

  outbDAC(value, DAC_PIXEL_MASK);
}

static uchar
inbDAC_CMD(void)
{
  int i;

  inbDAC(DAC_PW_INDEX);                /* reset the IPF                */

  for (i = 0; i < 4; i++)      /* set the IPF                  */
    inbDAC(DAC_PIXEL_MASK);

  return inbDAC(DAC_PIXEL_MASK);
}


/*\
|*|    true color CLUT operations
\*/

static void
set_HC_LUT_entries(ggi_palette *table)
{
  int len;
  ggi_color *data;

  if (table && (data = table->data) && (len = table->cnt))
    {
      outbDAC(table->first >> 2, DAC_PW_INDEX);
      while (len > 0)
       {
         outbDAC(data->r >> SHIFT, DAC_P_DATA);
         outbDAC(data->g >> SHIFT, DAC_P_DATA);
         outbDAC(data->b >> SHIFT, DAC_P_DATA);
         data += 4;
         len  -= 4;
       }
    }
}

static int
get_HC_LUT_entries(ggi_palette *table, unsigned char index, int cnt)
{
  int len = 0;
  ggi_color *data;

  if (table && (data = table->data) && table->cnt)
    {
      index &= 0xFC;

      len = (cnt > table->cnt) ? cnt : table->cnt;
      len -= (4 - (index & 3)) & 3;

      outbDAC((table->first = index) >> 2, DAC_PR_INDEX);

      while (len > 0)
       {
         int i;
         register ushort
         col = inbDAC(DAC_P_DATA);     data->r = COLOR(col);
         col = inbDAC(DAC_P_DATA);     data->g = COLOR(col);
         col = inbDAC(DAC_P_DATA);     data->b = COLOR(col);

         for (i = 0; i < 4; i++)
           data[i] = *data;

         data += 4;
         len  -= 4;
       }
    }
  return len;
}

/*\
|*|	GGI interface
\*/

int
kgi_ramdac_init(void)
{
  uchar pix = inbDAC(DAC_PIXEL_MASK), cmd;

  outbDAC(0xFF, DAC_PIXEL_MASK);

  cmd = inbDAC_CMD();
  outbDAC_CMD(cmd | DAC_CMD_ERPF);

  if (inbEXT(DAC_EXT_ID0) == 'S')
    if (((inbEXT(DAC_EXT_ID1) << 8) | inbEXT(DAC_EXT_ID2)) == 15025)
      {
        TRACE(printk("  %s %s rev %c dac driver v"DAC_VERSION"\n", \
              ramdac->manufact, ramdac->model, inbEXT(DAC_EXT_ID3)));
        outbEXT(DAC_AUX_8BIT, DAC_EXT_AUXCTRL);
	outbDAC(cmd, DAC_E_COMMAND);
	return EOK;
      }

  outbDAC(pix, DAC_PIXEL_MASK);			/* restore pixelmask	*/
  outbDAC(inbDAC(DAC_PW_INDEX), DAC_PW_INDEX);	/* reset adress counter	*/

  printk(" %s %s not detected.\n", ramdac->manufact, ramdac->model);
  return -ERAMDAC;
}

void
kgi_ramdac_done(void)
{
  TRACE(printk("%s %s driver removed.\n", ramdac->manufact, ramdac->model));
}

struct kgi_ramdac
*kgi_ramdac_get_info (void)
{
  return ramdac; 
}

int
kgi_ramdac_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int maxpclk, pclk;
  
  switch(mode->request.graphtype)
   {
     case GT_TEXT: 
     case GT_1BIT: 
     case GT_4BIT:	/* these all use 8bit PB */
     case GT_4BIT_PLANES:
     case GT_8BIT:
	switch (cmd)
	 {
	   case CMD_PROPOSE:
		mode->lclk.mul = mode->clk.mul = 1;
		mode->lclk.div = mode->clk.div = 1;
		break;

	   case CMD_CHECK:
		if (   (mode->lclk.mul != 1) || (mode->lclk.div != 1)
		    || (mode->clk.mul  != 1) || (mode->clk.div  != 1))
		  return -ERAMDAC;
		break;

	   default:
	 }
	maxpclk = ramdac->normal.max;
	break;

     case GT_15BIT:
     case GT_16BIT:
	switch(cmd)
	 {
	   case CMD_PROPOSE:
		mode->lclk.mul = mode->clk.mul = 1;
		mode->lclk.div = mode->clk.div = 1;
		break;

	   case CMD_CHECK:
		if (   (mode->lclk.div != 1)
		    || (mode->lclk.mul != mode->clk.mul)
		    || (mode->lclk.div != mode->clk.div))
		  return -ERAMDAC;
		break;

           default:
	 }
	switch(mode->lclk.mul)
	 {
	   case 1: maxpclk = ramdac->dblclk.max; break;
	   case 2: maxpclk = ramdac->normal.max; break;
	   default: return -ERAMDAC;
	 }
	break;

     case GT_24BIT:
	switch(cmd)
	 {
	   case CMD_PROPOSE:
		mode->lclk.div = mode->clk.div = 1;
		mode->lclk.mul = mode->clk.mul = 3;
		break;

	   case CMD_CHECK:
		if (   (mode->lclk.div != 1) || (mode->clk.div != 1)
		    || (mode->lclk.mul != 3) || (mode->clk.mul != 3))
		  return -ERAMDAC;
		break;

	   default:
         }
	maxpclk = ramdac->normal.max;
	break;

     case GT_32BIT:
	switch(cmd)
	 {
	   case CMD_PROPOSE:
		mode->lclk.mul = mode->clk.mul = 2;
		mode->lclk.div = mode->clk.div = 1;
		break;

	   case CMD_CHECK:
		if (   (mode->lclk.div != 1)
		    || (mode->lclk.mul != mode->clk.mul) 
		    || (mode->lclk.div != mode->clk.div))
		  return -ERAMDAC;
		break;

	   default:	
	 }
	switch(mode->lclk.div)
	 {
	   case 2:  maxpclk = ramdac->dblclk.max; break;
	   case 4:  maxpclk = ramdac->normal.max; break;
	   default:
		return -ERAMDAC;
	 }
	break;

     default:
	INTERNALERROR;
	return -ERAMDAC;
   }

  pclk = mode->dclk * mode->lclk.mul / mode->lclk.div;

  TRACE(printk("%s:%i: maxpclk = %i Hz, pclk = %i Hz\n", __FILE__, __LINE__, \
	maxpclk, pclk));

  switch(cmd)
   {
     case CMD_LOWER:
	if (maxpclk < pclk)
	  mode->dclk = maxpclk * mode->lclk.div / mode->lclk.mul;
	return EOK;

     case CMD_RAISE:
	return EOK;

     case CMD_CHECK:
	return ((pclk < 0) || (maxpclk < pclk)) ? -ERAMDAC : EOK;
     default:
	return EOK;
   }
}

void
kgi_ramdac_set_mode(struct kgi_mode *mode)
{
  uchar cmd = DAC_CMD_USECLUT,		/* command register value	*/
        rep = DAC_REP_MODE0;		/* repack register value	*/

  switch(mode->request.graphtype)
   {
     case GT_TEXT:
     case GT_1BIT:
     case GT_4BIT: 
     case GT_4BIT_PLANES: 
     case GT_8BIT:
		cmd = DAC_CMD_MODE0 /* ??? | DAC_C_NORMALCLK */;
		break;

     case GT_15BIT:
		cmd |= DAC_CMD_MODE1;
		cmd |= (mode->lclk.mul==1) ? DAC_CMD_DBL_CLK : DAC_CMD_NRM_CLK;
		break;

     case GT_16BIT:
		cmd |= DAC_CMD_MODE3;
		cmd |= (mode->lclk.mul==1) ? DAC_CMD_DBL_CLK : DAC_CMD_NRM_CLK;
		break;

     case GT_24BIT:
		cmd |= DAC_CMD_MODE4;
		break;

     case GT_32BIT:
		rep = DAC_REP_MODE1;
		cmd |= DAC_CMD_MODE4;
		cmd |= (mode->lclk.mul==2) ? DAC_CMD_DBL_CLK : DAC_CMD_NRM_CLK;
		break;
     default:
		INTERNALERROR;
		return;
   }

  TRACE(printk("%s:%i: %s mode, cmd = %.2X, rep = %.2X\n", __FILE__,__LINE__, \
               GRAPHMASK(mode->request.graphtype), cmd, rep));

  outbDAC_CMD(cmd | DAC_CMD_ERPF);	/* map extended registers	*/

  outbEXT(DAC_AUX_8BIT, DAC_EXT_AUXCTRL);

  outbEXT(0xFF, DAC_EXT_PIX_MASK0);
  outbEXT(0xFF, DAC_EXT_PIX_MASK1);
  outbDAC(0xFF, DAC_EXT_PIX_MASK2);

  outbEXT(rep, DAC_EXT_REPACK);
  outbDAC(cmd, DAC_E_COMMAND);		/* unmap extended registers	*/
  outbDAC(0xFF, DAC_PIXEL_MASK);

  if (  (mode->request.graphtype == GT_15BIT)
      ||(mode->request.graphtype == GT_16BIT))
    set_HC_LUT_entries(mode->palette);
  else
    set_CLUT_entries(mode->palette);
}

int
kgi_ramdac_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  struct kgi_state *gr = file->private_data;

  switch(cmd)
  { case RAMDAC_SETCLUT: 
    {  
           ggi_palette *h = (void*) _io_buf;
           int first,cnt;

           first=h->first;cnt=h->cnt;
           if (first+cnt > 256)
             return -ENBOGUS;

           if (0==verify_area(VERIFY_READ,(void *)h->data,sizeof(ggi_color)*cnt)) { 
             memcpy_fromfs (&(gr->color[first]),
                            (void *) h->data, sizeof(ggi_color)*cnt); 
           }
          if (  (gr->mode.request.graphtype == GT_15BIT)
               ||(gr->mode.request.graphtype == GT_16BIT))
             set_HC_LUT_entries(gr->mode.palette);
            else
            set_CLUT_entries(gr->mode.palette); /* set clut of mode */
    } return 0;
    case RAMDAC_GETCLUT: 
    {	   int x=0;
           ggi_palette *h = (void*) _io_buf;
           int first,cnt;
           _out_buf=_io_buf;

           first=h->first;cnt=h->cnt;
           if (first+cnt > 256)
             return -ENBOGUS;

           if (0==verify_area(VERIFY_WRITE,(void *)h->data,sizeof(ggi_color)*cnt)) { 
             memcpy_tofs ((void *) h->data,
             		  &(gr->color[first]),sizeof(ggi_color)*cnt); 
           }
        }
        return 0;
  }
  return -ENODRVSUP_ALWAYS_CANT;       /* This driver can't do this. */
}
