/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _LIB_COMPILE_

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <string.h>
#include <ggi/graphlib.h>


/*************************************/
/* fill (erase) the (virtual) screen */
/*************************************/

static int fillscreen_lower(int unused,void *unused2)
{ return(ggi_drawbox(0,  0,CurrentMode.virtual.x,CurrentMode.virtual.y)); }

static int fillscreen1bit(int unused,void *unused2)
{ memset(ggi_membuffer,(ggi_drwinfo.fg_color!=0),
	 CurrentMode.virtual.x*CurrentMode.virtual.y>>3);
  return(0);
}

static int fillscreen4bit(int unused,void *unused2)
{ memset(ggi_membuffer,
	 (ggi_drwinfo.fg_color << 4) | ggi_drwinfo.fg_color,
	 CurrentMode.virtual.x*CurrentMode.virtual.y>>1);
  return(0);
}

static int fillscreen4bit_pl(int unused,void *unused2)
{ set16col(ggi_drwinfo.fg_color);
  memset(ggi_membuffer,-1,CurrentMode.virtual.x*CurrentMode.virtual.y>>3);
  return(0);
}


static int fillscreen8bit(int unused,void *unused2)
{ memset(ggi_membuffer,ggi_drwinfo.fg_color,CurrentMode.virtual.x*CurrentMode.virtual.y); 
  return(0);
}

static int fillscreen16bit(int unused,void *unused2)
{ short *buf=(short *)ggi_membuffer;
  int len=CurrentMode.virtual.x*CurrentMode.virtual.y;
  while(len--) *buf++=ggi_drwinfo.fg_color;
  return(0);
}

static int fillscreen24bit(int unused,void *unused2)
{ int col[3];
  int *buf=(int *)ggi_membuffer;
  int len=CurrentMode.virtual.x*CurrentMode.virtual.y>>2,i;
  unsigned char *colp=(unsigned char *)col;
  for (i=0;i<4;i++)
    {
      *colp++=(ggi_drwinfo.fg_color)&0xFF;
      *colp++=(ggi_drwinfo.fg_color>>8)&0xFF;
      *colp++=(ggi_drwinfo.fg_color>>16)&0xFF;
    };
  while(len--) 
    {
      *buf++=col[0];
      *buf++=col[1];
      *buf++=col[2];
    };
  return(0);
}

static int fillscreen32bit(int unused,void *unused2)
{ int *buf=(int *)ggi_membuffer;
  int len=CurrentMode.virtual.x*CurrentMode.virtual.y;
  while(len--) *buf++=ggi_drwinfo.fg_color;
  return(0);
}

int ggi_fillscreen(void)
{
	switchavail(GRAPHICS_FILLSCREEN,NULL);

	switcha(GRAPHICS_FILLSCREEN)
	{ case AVM_MMAP :
		switch(CurrentMode.graphtype)
		  { case GT_1BIT:setdoemu(GRAPHICS_FILLSCREEN,AVE_MODE,fillscreen1bit,NULL);
		  case GT_4BIT:setdoemu(GRAPHICS_FILLSCREEN,AVE_MODE,fillscreen4bit,NULL);
		  case GT_4BIT_PLANES:setdoemu(GRAPHICS_FILLSCREEN,AVE_MODE,fillscreen4bit_pl,NULL);
		  case GT_8BIT:setdoemu(GRAPHICS_FILLSCREEN,AVE_MODE,fillscreen8bit,NULL);
		  case GT_16BIT:setdoemu(GRAPHICS_FILLSCREEN,AVE_MODE,fillscreen16bit,NULL);
		  case GT_24BIT:setdoemu(GRAPHICS_FILLSCREEN,AVE_MODE,fillscreen24bit,NULL);
		  case GT_32BIT:setdoemu(GRAPHICS_FILLSCREEN,AVE_MODE,fillscreen32bit,NULL);
  		  default     :break;}
  		/* Fallthrough */
	  case AVM_LOWER:
	  	setdoemu(GRAPHICS_FILLSCREEN,AVE_ALWAYS,fillscreen_lower,NULL);
	  	break;
	  default       :
	  	ggi_panic("fillscreen - no method.");break;
	}
	return(0);
}
