/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _LIB_COMPILE_

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <string.h>
#include <ggi/graphlib.h>


/********************************/
/* set any mode (text/graphics) */
/********************************/
int ggi_setmode(ggi_mode *tm)
{ int x;
  /* Delete ioctls marked as "unsupported this mode" */
  for(x=0;x<sizeof(ioctl_avail)/sizeof(ioctl_avail[0]);x++) 
    if ((ioctl_avail[x]&AVE_MASK)==AVE_MODE) 
    { ioctl_avail[x]=AVM_IOCTL|AVE_TEST;
      emu_func[x]=ggi_ioctl; }

  ioctl(ggi_tty_fd, KDSETMODE, KD_GRAPHICS);	/* We probably go to graphics now ... */

  if ((x=ggi_ioctl(DRIVER_SETGRAPHMODE,tm))>=0) 
  { CurrentMode=*tm;
    if (tm->graphtype==GT_TEXT)
    { struct vt_sizes vts;
      vts.v_rows = tm->visible.y/tm->textgrid.y;
      vts.v_cols = tm->visible.x/8;
      vts.v_scrollsize = 0;
      if (ioctl(ggi_tty_fd, VT_RESIZE, &vts)) perror("VT_RESIZE");
      ioctl(ggi_tty_fd, KDSETMODE, KD_TEXT);
/*      system("setfont");	not needed any more */	
    } else {
      ggi_setcolor(0);
      ggi_fillscreen();
    }
  }

  return(x);
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int ggi_checkmode(ggi_mode *tm)
{ return(ggi_ioctl(DRIVER_TESTMODE,tm)); }

/************************/
/* get the current mode */
/************************/
int ggi_getmode(ggi_mode *tm)
{ return(ggi_ioctl(DRIVER_GETGRAPHMODE,&tm)); }

/**************************************************/
/* keep the current mode after closing the device */
/**************************************************/
int ggi_keepmode(void)
{ return(ggi_ioctl(DRIVER_KEEPMODE,NULL)); }

/******************/
/* set a textmode */
/******************/
int ggi_textmode(int cols,int rows,int fontsizex,int fontsizey)
{ ggi_mode tim={	{800,600},	/* Default size */
        	        {1024,1024},
        	        {300,220},
                        GT_TEXT,
                	{8,16},
                	/*0*/           };
                        
  if (fontsizex<4||fontsizex>32) return -EINVAL;	/* Maybe some card can do it ... */
  if (fontsizey<4||fontsizey>32) return -EINVAL;

  tim.virtual.x=tim.visible.x=cols*fontsizex;
  tim.virtual.y=tim.visible.y=rows*fontsizey;
  tim.textgrid.x=fontsizex;
  tim.textgrid.y=fontsizey;

  return(ggi_setmode(&tim));
}

/***********************/
/* set a graphics mode */
/***********************/
int ggi_graphmode(int xsize,int ysize,int xvirtual,int yvirtual,int type)
{ ggi_mode tim={ { 800, 600},
	         {1024,1024},
	         { 300, 220},
	         GT_8BIT,
	         {8,16},
	         /*0*/      };
  tim.visible.x=xsize;
  tim.visible.y=ysize;
  tim.virtual.x=xvirtual;
  tim.virtual.y=yvirtual;                       
  tim.graphtype=type;

  return(ggi_setmode(&tim));
}

/*************************/
/* check a graphics mode */
/*************************/
int ggi_checkgraphmode(int xsize,int ysize,int xvirtual,int yvirtual,int type)
{ ggi_mode tim={ { 800, 600},
	         {1024,1024},
	         { 300, 220},
	         GT_8BIT,
	         {8,16},
	         /*0*/      };
  tim.visible.x=xsize;
  tim.visible.y=ysize;
  tim.virtual.x=xvirtual;
  tim.virtual.y=yvirtual;                       
  tim.graphtype=type;

  return(ggi_checkmode(&tim));
}

